import yaml
import logging
import re

from collections import namedtuple
from werkzeug.exceptions import BadRequest
from ingaia.commons import request_util
from decorator import decorator
from .exceptions import NoOpenAPIFileConfiguredException


class ApiConfig:
    """
    Class for work with open api config properties
    """
    _api_info = None
    _config_file = None

    @classmethod
    def set_config_file(cls, openapi_file):
        cls._config_file = openapi_file

    @classmethod
    def load_config_file(cls):
        """
        Loads the api config file
        :return:
        """
        if cls._api_info is None:
            if cls._config_file is None:
                raise NoOpenAPIFileConfiguredException()

            # reads the yaml file
            with open(cls._config_file, 'r') as stream:
                try:
                    api_info = dict(yaml.safe_load(stream))
                    logging.info(f' The file "{cls._config_file}" was successfully loaded')
                except yaml.YAMLError as exc:
                    raise exc
            # convert the dict to an object to export
            cls._api_info = namedtuple("ApiInfo", api_info.keys())(*api_info.values())

    @classmethod
    def config(cls):
        """
        Returns the dict with api configurations
        :return:
        """
        if cls._api_info is None:
            cls.load_config_file()
        return cls._api_info

    @classmethod
    def get_params_by_request_body(cls, request_body_def):
        """
        Gets the request body params based on openapi.yaml definitions
        :param request_body_def: Definition name from yaml file
        :return: Dict for params
        """
        # load the request body parameters
        list_subscription = cls.config().components['requestBodies'][request_body_def]
        list_subscription_params = list_subscription['content']['application/json']['schema']['properties']

        # make the response
        default_params = []
        operators_params = []
        for param in list_subscription_params:
            operator = cls.get_param_operator(param)
            if not operator:
                default_params.append(param)
            else:
                param_value = list_subscription_params[param]
                operator_param = {
                    "param": str(param).replace(operator, '').replace('[]', ''),
                    "enum": param_value["enum"] if "enum" in param_value else None
                }

                # gets supported opertator params
                description = str(param_value['description'])

                # NOTE: for this to work, it's necessary begin the description
                #       with "**Supported params**:" and finish the 'params' with "<br>"
                # Syntax example:
                #   **Supported params**: asc, desc <br>
                supported_params = '**Supported params**:'
                if description.find(supported_params) != 0:
                    logging.warning(f'Description error for "{param}"')
                else:
                    if description.find('<br>') != -1:
                        operators = description[len(supported_params):description.find('<br>')].split(',')
                        operators = [o.strip() for o in operators]
                        operator_param["operators"] = operators

                operators_params.append(operator_param)

        return {
            "default": default_params,
            "operators": operators_params
        }

    @classmethod
    def validate_payload(cls, payload, request_body_def):
        """
        Validates the request body (payload) based on openapi.yaml definitions
        :param payload: Payload received from the request
        :param request_body_def: Definition name from yaml file
        :raise BadRequest if the request body is invalid
        """
        request_body_params = cls.get_params_by_request_body(request_body_def)

        for param in payload.keys():
            in_default_params = True
            if param not in request_body_params['default']:
                in_default_params = False

            in_operators_params = False
            if not in_default_params:
                operator: str = cls.get_param_operator(param)
                if operator:
                    in_operators_params = True
                    cls._validade_operator(param, operator, payload[param], request_body_params)

            if not in_default_params and not in_operators_params:
                raise BadRequest(f'Invalid param "{param}"')

    @classmethod
    def _validade_operator(cls, param, operator, value, request_body_params):
        """
        Validates the param with opertator
        :param operator: Operator received in payload
        :param value: Value received in payload
        :param param: Parameter received in payload
        :param request_body_params: Parameters loaded from api config file
        """
        param = str(param).replace(operator, '').replace('[]', '')
        if 'operators' in request_body_params:
            param_operators = [p["param"] for p in request_body_params['operators']]
            if param in param_operators:
                valid_operators = []
                valid_values = []
                for p in request_body_params['operators']:
                    if p["param"] == param:
                        valid_operators = p["operators"] if "operators" in p else None
                        valid_values = p["enum"] if "enum" in p else None
                        break

                if valid_operators and operator not in valid_operators:
                    raise BadRequest(f'Invalid operator "{operator}" for parameter "{param}"')

                if valid_values and value not in valid_values:
                    raise BadRequest(f'Invalid value "{value}" for parameter "{param}". Valid values: {valid_values}')

    @staticmethod
    def get_param_operator(param) -> str:
        """
        Auxiliary method to receive the operator of a param
        :param param:
        :return:
        """
        matches = re.findall(r'\[(.*?)\]', param)
        return matches[0] if matches else None


@decorator
def validate_payload(fn, request_body_def=None, *args, **kw):
    # recovers the payload
    payload = dict(request_util.get_request_payload())

    # aditional validation
    # validates the payload based on api config file (openapi.yaml)
    ApiConfig.validate_payload(payload, request_body_def)

    # returns the function
    return fn(*args, **kw)



import os
import jwt

from jwt.exceptions import PyJWTError
from werkzeug.exceptions import Unauthorized

jwt_key_pub_path = f'{os.path.abspath(os.path.dirname(__file__))}/jwt-key.pub'


def decode_token(token):
    """
    Function responsible for decode the token informed by the client

    :param token:   JWT Token
    :return:        json decoded
    :raise:         Unauthorized when token is invalid
    """
    try:
        return jwt.decode(token, open(jwt_key_pub_path).read(), algorithms=['RS256'])
    except PyJWTError as e:
        raise Unauthorized()



import json
from flask import request
from werkzeug.exceptions import BadRequest


def get_request_payload():
    if request.data is None or request.data == b'':
        raise BadRequest('The request body is empty')

    payload = json.loads(request.data)

    if payload is None:
        raise BadRequest('The payload is empty')

    return payload



import logging

from google.cloud import storage
from google.api_core.exceptions import NotFound
from google.cloud.storage.bucket import Bucket
from .exceptions import NoBlobSetException
from .project import ProjectReference


class StorageUtil:
    """Auxiliary class for Google Cloud Storage"""

    def __init__(self, bucket_name, file_name=None, project=None, project_id=None, location=None):
        self.client = storage.Client()
        self.project = project if project else ProjectReference(project_id, location)
        self._bucket_name = bucket_name
        self._bucket = None
        self._blob = None
        self._file_name = file_name
        if file_name:
            self.set_blob(file_name)

    @property
    def bucket(self):
        if not self._bucket:
            try:
                self._bucket = self.client.get_bucket(self._bucket_name)
            except NotFound:
                self._bucket = Bucket(client=self.client, name=self._bucket_name)
                self._bucket.create(client=self.client, location=self.project.location)
                logging.info('Bucket {} not found and was created.'.format(self._bucket.name))

        return self._bucket

    @property
    def blob(self):
        if not self._blob:
            raise NoBlobSetException()
        return self._blob

    @property
    def file_name(self):
        if not self._blob:
            raise NoBlobSetException()
        return self._file_name

    def set_blob(self, file_name):
        self._file_name = file_name
        self._blob = self.bucket.get_blob(file_name)
        if not self._blob:
            self._blob = self._bucket.blob(file_name)
            logging.info(f'File not found and was created: {file_name}')
        return self

    def delete_blob(self):
        try:
            self.bucket.delete_blob(self.file_name)
            logging.info(f'File deleted: {self.file_name}')
        except NotFound:
            logging.info(f'File not found: {self.file_name}')

        self._file_name = None
        self._blob = None
        return self

    def new_content(self, new_content: str):
        content_len = 80
        content = f'{new_content[:content_len]} [...]' if len(new_content) > content_len else new_content
        logging.info(f'Uploading new content to file "{self._file_name}": {repr(content)}')
        self.blob.upload_from_string(new_content)
        return self

    def get_content(self):
        return self.blob.download_as_string()

    @classmethod
    def buckets(cls):
        return [b for b in cls.client.list_buckets()]



from google.cloud import scheduler
from .project import ProjectReference
from .exceptions import ProjectReferenceNotDefinedException


class SchedulerUtil:
    """
    Auxiliary class for Google Cloud Scheduler
    """
    CRON_HEADERS = ['X-Appengine-User-Ip', 'X-Appengine-Cron']

    def __init__(self, project=None, project_id=None, location=None):
        self.client = scheduler.CloudSchedulerClient()
        self.project = project if project else ProjectReference(project_id, location)

    def create_job(self, job_def: dict):
        """
        Creates a cron job in Google Cloud Scheduler

        :param job_def: Job defintion. See the link below:
        https://googleapis.dev/python/cloudscheduler/latest/gapic/v1/types.html#google.cloud.scheduler_v1.types.Job
        :raises ProjectReferenceNotDefinedException
        """
        if not self.project:
            raise ProjectReferenceNotDefinedException()

        parent = self.client.location_path(self.project.project_id, self.project.location)

        self.client.create_job(parent, job_def)

    @classmethod
    def is_cron_request(cls, request):
        """
        Verify if a flask request is generated by GCloud Cron Service

        "The X-Appengine-Cron header is set internally by Google App Engine. If your request handler finds
        this header it can trust that the request is a cron request. The X- headers are stripped by App Engine
        when they originate from external sources so that you can trust this header."

        Reference:
        https://cloud.google.com/appengine/docs/standard/python3/scheduling-jobs-with-cron-yaml#validating_cron_requests

        :param request: Flask Response object
        :return:        If the request was created by a google cloud cron job
        """
        headers = dict(t for t in request.headers if t[0] in cls.CRON_HEADERS)

        if set(cls.CRON_HEADERS) == headers.keys():
            return headers['X-Appengine-Cron']

        return False



from google.cloud import datastore
from .exceptions import QueryIsNoneException


class DataStoreUtil:
    """
    Auxiliary class for Google Cloud Scheduler
    TODO it needs to be improved
    """

    def __init__(self, kind):
        self._kind = kind
        self._client = datastore.Client()
        self._entity = None
        self._query = None

    @property
    def entity(self):
        if not self._entity:
            self._entity = datastore.Entity(key=self._client.key(self._kind))
        return self._entity

    def put(self, entry):
        # load the entity
        self.entity.update(entry)
        # save in datastore
        self._client.put(self.entity)
        return self.entity

    @property
    def query(self):
        if not self._query:
            self._query = self._client.query(kind=self._kind)
        return self._query

    def query_reset(self):
        self._query = None

    def delete_multi(self, query=None):
        if not query:
            raise QueryIsNoneException()

        keys = [self._client.key(self._kind, e.id) for e in query.fetch()]

        self._client.delete_multi(keys)



import logging
from google.cloud import firestore


class FireStoreUtil:
    """
    Auxiliary class for Google Cloud Scheduler

    Example:
    - Save a document:
        FireStoreUtil('coll', 'doc_id').set({...})
    - Select a document:
        FireStoreUtil('coll', 'doc_id').get()
    - Query:
        coll = FireStoreUtil('coll').collection
        docs = coll.where(u'field', u'==', u'value').limit(1).stream()
    """

    def __init__(self, collection_name, document_name=None):
        self._collection_name = collection_name
        self._document_name = document_name
        self._client = firestore.Client()
        self._collection = None
        self._document = None

    @property
    def collection(self):
        if not self._collection:
            self._document = self._client.collection(self._collection_name)
        return self._document

    @property
    def document(self, document_name=None):
        # forces new document
        if document_name is not None:
            self._document = self.collection.document(document_name)
        # creates a new document if it doesnt exist
        if not self._document:
            self._document = self.collection.document(self._document_name)
        return self._document

    def get(self):
        return self.document.get()

    def set(self, document):
        return self.document.set(document)

    def set_document_name(self, document_name):
        self._document_name = document_name

    def get_documents(self):
        return self.collection.stream()

    @staticmethod
    def delete_multi(docs):
        for doc in docs:
            logging.debug(u'Deleting doc {} => {}'.format(doc.id, doc.to_dict()))
            doc.reference.delete()

    def delete_collection(self, batch_size=10):
        docs = self.collection.limit(batch_size).stream()
        deleted = 0

        for doc in docs:
            logging.debug(u'Deleting doc {} => {}'.format(doc.id, doc.to_dict()))
            doc.reference.delete()
            deleted = deleted + 1

        if deleted >= batch_size:
            return self.delete_collection(batch_size)
